using System;
using LinearAlgebra;

namespace DerivativeFunctions
{
	/// <summary>
	/// Represents the right hand side of a differential equation.
	/// </summary>
	public interface DerivativeFunction
	{
		VectorN Evaluate(double t, VectorN v);
		MatrixN Gradient(double t, VectorN v);
		int MinimumDimensions {get;}
	}

	public class HarmonicOscillator : DerivativeFunction
	{
		double mass;
		double strength;

		public HarmonicOscillator(double mass, double strength)
		{
			this.mass = mass;
			this.strength = strength;
		}

		public int MinimumDimensions
		{
			get
			{
				return 2;
			}
		}

		public VectorN Evaluate(double t, VectorN v)
		{
			int d = v.Size;
			if(d % 2 != 0)
			{
				throw new Exception("Number of dimensions has to be even.");
			}

			VectorN result = new VectorN(d);
			for(int i = 0; i < d/2; i++)
			{
				result.Set(2*i, v.Get(2*i+1)/mass);
				result.Set(2*i+1, -strength*v.Get(2*i));
			}

			return result;
		}

		public MatrixN Gradient(double t, VectorN v)
		{
			if(v.Size % 2 != 0)
			{
				throw new Exception("Number of dimensions has to be even.");
			}
			if(v.Size != 2)
			{
				throw new Exception("Not implemented.");
			}

			double[,] m = new double[,] {{0.0, 1.0/mass}, {-strength, 0.0}};
			return new MatrixN(ref m);
		}
	}

	public class ExponentialGrowth : DerivativeFunction
	{
		double exponent;

		public ExponentialGrowth(double exponent)
		{
			this.exponent = exponent;
		}

		public int MinimumDimensions
		{
			get
			{
				return 1;
			}
		}

		public VectorN Evaluate(double t, VectorN v)
		{
			int d = v.Size;
			VectorN result = new VectorN(d);
			for(int i = 0; i < d; i++)
			{
				result.Set(i, v.Get(i)*exponent);
			}

			return result;
		}

		public MatrixN Gradient(double t, VectorN v)
		{
			MatrixN m = new MatrixN(v.Size);
			for(int i = 0; i < v.Size; i++)
			{
				m.Set(i, i, exponent);
			}
			return m;
		}
	}

	public class MatrixMultiplication : DerivativeFunction
	{
		private MatrixN m;

		public MatrixMultiplication(ref MatrixN m)
		{
			this.m = m;
		}

		public int MinimumDimensions
		{
			get
			{
				return 1;
			}
		}

		public VectorN Evaluate(double t, VectorN v)
		{
			if(v.Size != m.Size)
			{
				throw new Exception("Dimension mismatch.");
			}
			return m*v;
		}

		public MatrixN Gradient(double t, VectorN v)
		{
			return m;
		}
	}

	public class StrangeInstabilityEffect : DerivativeFunction
	{
		public StrangeInstabilityEffect()
		{}

		public int MinimumDimensions
		{
			get
			{
				return 1;
			}
		}

		public VectorN Evaluate(double t, VectorN v)
		{
			int d = v.Size;
			VectorN result = new VectorN(d);
			for(int i = 0; i < d; i++)
			{
				result.Set(i, t - v.Get(i)*v.Get(i));
			}

			return result;
		}

		public MatrixN Gradient(double t, VectorN v)
		{
			MatrixN m = new MatrixN(v.Size);
			for(int i = 0; i < v.Size; i++)
			{
				m.Set(i, i, -2.0*v.Get(i));
			}
			return m;
		}
	}

	public class Lorenz : DerivativeFunction
	{
		public Lorenz()
		{}

		public int MinimumDimensions
		{
			get
			{
				return 3;
			}
		}

		public VectorN Evaluate(double t, VectorN v)
		{
			return new VectorN( 10.0*(v.Get(1)-v.Get(0)),
								28.0*v.Get(0) - v.Get(1) - v.Get(0)*v.Get(2),
								-(8.0/3.0)*v.Get(2) + v.Get(0)*v.Get(1) );
		}

		public MatrixN Gradient(double t, VectorN v)
		{
			throw new Exception("Not implemented.");
		}
	}

	public class Rssler : DerivativeFunction
	{
		public Rssler()
		{}

		public int MinimumDimensions
		{
			get
			{
				return 3;
			}
		}

		public VectorN Evaluate(double t, VectorN v)
		{
			return new VectorN( -v.Get(1) - v.Get(2),
								v.Get(0) + 0.2*v.Get(1),
								0.2 + (v.Get(0) - 5.7)*v.Get(2) );
		}

		public MatrixN Gradient(double t, VectorN v)
		{
			throw new Exception("Not implemented.");
		}
	}

	public class SpringPendulum2D : DerivativeFunction
	{
		double restLength;
		double mass;
		double strength;
		double gravity;
		double damping;

		public SpringPendulum2D(double restLength, double mass, double strength, double gravity, double damping)
		{
			this.restLength = restLength;
			this.mass = mass;
			this.strength = strength;
			this.gravity = gravity;
			this.damping = damping;
		}

		public int MinimumDimensions
		{
			get
			{
				return 4;
			}
		}

		public VectorN Evaluate(double t, VectorN v)
		{
			if(v.Size != 4)
			{
				throw new Exception("Dimension mismatch.");
			}
			double length = Math.Sqrt(v.Get(0)*v.Get(0) + v.Get(2)*v.Get(2));
			return new VectorN(v.Get(1)/mass, - strength*(length-1.0)*v.Get(0)/length - (damping/mass)*v.Get(1),
							   v.Get(3)/mass, - strength*(length-1.0)*v.Get(2)/length - (damping/mass)*v.Get(3) - gravity);
		}

		public MatrixN Gradient(double t, VectorN v)
		{
			throw new Exception("Not implemented");
		}
	}

	public class ReducedPrecision : DerivativeFunction
	{
		private DerivativeFunction f;
		private double relativeError;
		private Random rand;

		public ReducedPrecision (DerivativeFunction f, double relativeError)
		{
			this.f = f;
			this.relativeError = relativeError;
			rand = new Random();
		}

		public VectorN Evaluate(double t, VectorN v)
		{
			return f.Evaluate(t, v) * (1.0 + relativeError*(2.0*rand.NextDouble()-1.0));
		}

		public MatrixN Gradient(double t, VectorN v)
		{
			throw new Exception("Not implemented.");
		}

		public int MinimumDimensions
		{
			get
			{
				return f.MinimumDimensions;
			}
		}
	}
}
